/**
 * tkz-euclide 在 js 中的模仿。注意函数名称和用法与 tkz-euclide 略有区别，并且函数不够完全。
 * @author: maven
 * @version: 1.1
 * @copyleft
 */

/**
 * 点
 * @param name: 点名称
 * @param x: 横坐标
 * @param y: 纵坐标
 */
function tkzPoint(name, x, y) {
	this.name = name;
	this.x = parseFloat(x);
	this.y = parseFloat(y);
}
/**
 * tkz 画板
 * @param ctx: canvers 对象
 */
function tkz(ctx) {
	let it = this;
	/**
	 * 所有点的集合，以点名称为索引
	 */
	this.points = {};
	// pt 到 cm 的转换系数
	this.cm_pt = 37.8;
	/**
	 * 初始化 ctx。
	 * @param x: 原点在画布上的 x 坐标
	 * @param y: 原点在画布上的 y 坐标
	 */
	this.init = function(x,y) {
		ctx.translate(x,y);
		// 使 y 轴向上
		ctx.scale(1,-1);
	}
	/**
	 * 设置填充颜色、笔触颜色、笔画样式等参数
	 * @param params: 绘制参数
	 * @return: 返回设置后的绘制参数
	 */
	function setParams(params) {
		if(!params.color) {
			params.color = 'black';
		}
		if(!params.fill) {
			params.fill = 'black';
		}
		if(!params.width) {
			params.width = 1;
		}
		if(params.dash) {
			ctx.setLineDash([5, 5]);
		} else {
			ctx.setLineDash([]);
		}
		if(!params.font) {
			params.font = "14px Arial";
		}
		return params;
	}
	/**
	 * 判断数值是否为零
	 * @param number: 被判断的数值
	 * @return: 数值为零返回 true，否则返回 false
	 */
	function isZero(number) {
		return Math.abs(number) < 1E-8;
	}
	/**
	 * 过两点的直线的一般式，返回 Ax+By+C = 0 中的 A,B,C
	 * @param start: 起点
	 * @param to: 终点
	 * @return: 返回 Ax+By+C = 0 中的 A,B,C
	 */
	function getABC(start, to) {
		return {A: (to.y-start.y), B: (start.x-to.x), C: (to.x*start.y-start.x*to.y)}
	}

	/**
	 * 取得圆半径
	 * @param center: 圆心
	 * @param point: 圆上一点
	 * @return: 返回圆半径
	 */
	function getR(center, point) {
		return Math.sqrt((center.x-point.x) * (center.x-point.x) + (center.y-point.y) * (center.y-point.y));
	}

	/**
	 * 判断点是否在直线上
	 * @param one: 直线上一点
	 * @param two: 直线上另一点
	 * @param point: 被判断的点
	 * @return: 在直线上返回 true，否则返回 false
	 */
	function atLine(one, two, point) {
		let abc = getABC(one, two);
		return isZero(abc.A*point.x+abc.B*point.y+abc.C);
	}
	/**
	 * 判断点是否在圆周上
	 * @param center: 圆心
	 * @param point: 圆周上一点
	 * @param p: 被判断的点
	 */
	function atCircle(center, point, p) {
		let r2 = (center.x-point.x)*(center.x-point.x)+(center.y-point.y)*(center.y-point.y);
		let d2 = (center.x-p.x)*(center.x-p.x)+(center.y-p.y)*(center.y-p.y);
		return isZero(r2-d2);
	}
	/**
	 * 内部方法，创建一些点。用于其它方法取得交点后的链式调用，点坐标由临时变量取得
	 * @param points_name: 点序列名称，用逗号分隔，例如 'A,B'，将从内部创建 A,B 两个点
	 */
	function create_point(points_name) {
		points_name = points_name.split(',');
		for(let i = 0; i < points_name.length; i++) {
			let point_name = points_name[i];
			let xy = it.tkz_tmp_points[i];
			let point = new tkzPoint(point_name, xy.x, xy.y);
			console.log(point);
			it.points[point_name] = point;
		}
	}
	/**
	 * 创建点失败时调用此函数
	 * @param point_name: 创建失败的点名称
	 */
	function create_point_err(point_name) {
		console.log(`点 ${point_name} 不能被创建: ${it.tkz_msg}`);
	}
	/**
	 * 取得两点间的距离
	 * @param one_name: 第一点
	 * @param two_name: 第二点
	 * @param dir: 距离是否有向，坐标轴正方向为正向。默认为无向距离，当值为 true 时为有向距离
	 * @return: 返回两点间的距离
	 */
	this.tkzCalcLength = function(one_name, two_name, dir) {
		let one = this.points[one_name];
		let two = this.points[two_name];
		// 无向距离
		let d = Math.sqrt((one.x-two.x)*(one.x-two.x)+(one.y-two.y)*(one.y-two.y));
		if(dir) {
			if(isZero(one.x-two.x)) {
				// x 坐标相等
				d *= (two.y < one.y ? -1 : 1);
			} else {
				// x 坐标不等
				d *= (two.x < one.x ? -1 : 1);
			}
		}
		return d;
	}
	/**
	 * 定义点
	 * @param points_strs: 点的字符串，例如 '1/0/A,2/0/B' 将定义 A(1,0) 和 B(2,0) 两个点
	 */
	this.tkzDefPoints = function(points_strs) {
		let point_strs = points_strs.split(',');
		for(let point_str of point_strs) {
			let point_params = point_str.split('/');
			this.tkz_tmp_points = [{'x': point_params[0], 'y': point_params[1]}];
			create_point(point_params[2]);
		}
	}
	/**
	 * 创建两点的中点。使用方法：tkzDefMidPoint('A','B')('Y')，创建线段 AB 的中点 Y
	 * @param one_name: 第一点
	 * @param tow_name: 第二点
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzDefMidPoint = function(one_name, tow_name) {
		let one = this.points[one_name];
		let two = this.points[tow_name];
		let x = (one.x+two.x)/2;
		let y = (one.y+two.y)/2;
		this.tkz_tmp_points = [{'x': x, 'y': y}];
		return create_point;
	}
	/**
	 * 旋转点。使用方法：tkzDefRotaPoint('O','B',-45)('Z')，将点 B 绕点 O 顺时针旋转 45 度，得到点 Z。
	 * @param center_name: 旋转中心
	 * @param point_name: 被旋转的点
	 * @param angle: 旋转角
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzDefRotaPoint = function(center_name, point_name, angle) {
		let center = this.points[center_name];
		let point = this.points[point_name];
		let theta = -angle*Math.PI/180;
		let x = (point.x-center.x)*Math.cos(theta)+(point.y-center.y)*Math.sin(theta)+center.x;
		let y = -(point.x-center.x)*Math.sin(theta)+(point.y-center.y)*Math.cos(theta)+center.y;
		this.tkz_tmp_points = [{'x': x, 'y': y}];
		return create_point;
	}
	/**
	 * 过一点作直线的垂线，得到垂足点。使用方法：tkzDefProjPoint('A','B','C')('Z')，过 C 作 AB 的垂线，得到垂足点 Z。
	 * @param one_name: 直线上一点
	 * @param two_name: 直线上另一点
	 * @param point_name: 直线外一点
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzDefProjPoint = function(one_name, two_name, point_name) {
		let one = this.points[one_name];
		let two = this.points[two_name];
		let point = this.points[point_name];
		let abc = getABC(one, two);

		let d = abc.A*abc.A+abc.B*abc.B;
		let x = (abc.B*abc.B*point.x-abc.A*abc.B*point.y-abc.A*abc.C)/d;
		let y = (abc.A*abc.A*point.y-abc.A*abc.B*point.x-abc.B*abc.C)/d;
		this.tkz_tmp_points = [{'x': x, 'y': y}];
		return create_point;
	}
	/**
	 * 作一点关于另两点连线的对称点，得到对称点。使用方法：tkzDefReflPoint('A','B','C')(`C'`)，作 C 关于 AB 的对称点C'。
	 * @param one_name: 直线上一点
	 * @param two_name: 直线上另一点
	 * @param point_name: 被对称的点
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzDefReflPoint = function(one_name, two_name, point_name) {
		let one = this.points[one_name];
		let two = this.points[two_name];
		let point = this.points[point_name];
		let abc = getABC(one, two);
		let d = abc.A*abc.A+abc.B*abc.B;
		let x = ((abc.B*abc.B-abc.A*abc.A)*point.x-2*abc.A*(abc.B*point.y+abc.C))/d;
		let y = ((abc.A*abc.A-abc.B*abc.B)*point.y-2*abc.B*(abc.A*point.x+abc.C))/d;
		this.tkz_tmp_points = [{'x': x, 'y': y}];
		return create_point;
	}
	/**
	 * 定比分点。使用方法：tkzDefLinearPoint('A','B',2)('C')，定义点 C，使得 AC : CB = K。
	 * @param one_name: 直线上一点
	 * @param two_name: 直线上另一点
	 * @param K: 定比
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzDefLinearPoint = function(one_name, two_name, K) {
		if(isZero(K+1)) {
			// 禁止定比为 -1
			this.tkz_msg = '禁止定比为 -1'
			return create_point_err;
		}
		let one = this.points[one_name];
		let two = this.points[two_name];
		let x = (one.x+K*two.x)/(1+K);
		let y = (one.y+K*two.y)/(1+K);
		this.tkz_tmp_points = [{'x': x, 'y': y}];
		return create_point;
	}

	/**
	 * 过一点作直线的垂线，并在垂线上取一点。使用方法：tkzDefOrthLine('A','B','P')('C')，过 P 作 AB 的垂线，并在垂线上取一点 C。
	 * @param one_name: 直线上一点
	 * @param two_name: 直线上另一点
	 * @param point_name: 作垂线所过的点
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzDefOrthLine = function(one_name, two_name, point_name) {
		let one = this.points[one_name];
		let two = this.points[two_name];
		let point = this.points[point_name];
		if(!atLine(one, two, point)) {
			// 点 point 不在直线上，取得点到直线的垂足
			return this.tkzDefProjPoint(one_name, two_name, point_name);
		} else {
			// 点 point 在直线上，将点 one 绕 point 旋转 90 度
			return this.tkzDefRotaPoint(point_name, one_name, 90);
		}
	}
	/**
	 * 过一点作直线的平行线，并在平行线上取一点。使用方法：tkzDefParaLine('A','B','P')('C')，过 P 作 AB 的平行线，并在平行线上取一点 C。
	 * @param one_name: 直线上一点
	 * @param two_name: 直线上另一点
	 * @param point_name: 作平行线所过的点
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzDefParaLine = function(one_name, two_name, point_name) {
		let one = this.points[one_name];
		let two = this.points[two_name];
		let point = this.points[point_name];
		// 将点 two 绕点 one 旋转 90 度，得到点 tkz_tmp_p
		this.tkzDefRotaPoint(one_name, two_name, 90)('tkz_tmp_X');
		let X = this.points['tkz_tmp_X'];
		if(!atLine(one, X, point)) {
			// 点 point 不在直线 one-X 上，取得点到直线的垂足
			return this.tkzDefProjPoint(one_name, 'tkz_tmp_X', point_name);
		} else {
			// 点 point 在直线 one-tkz_tmp_p 上
			// 将点 one 绕点 two 旋转 90 度，得到点 tkz_tmp_p
			this.tkzDefRotaPoint(two_name, one_name, 90)('tkz_tmp_X');
			// 取得点到直线的垂足
			return this.tkzDefProjPoint(two_name, 'tkz_tmp_X', point_name);
		}
	}
	/**
	 * 求直线到直线的角。使用方法：tkzGetAngle('P1','P2','T1','T2')，得到 T1T2 到 P1P2 的角
	 */
	this.tkzGetAngle = function(P1,P2,T1,T2) {
		let l1_abc = getABC(this.points[P1],this.points[P2]);
		let l2_abc = getABC(this.points[T1],this.points[T2]);
		// 到角公式求 T1T2 到 P1P2 的角
		let d = l1_abc.A*l2_abc.A+l1_abc.B*l2_abc.B;
		let theta = 90;
		if(l1_abc.A*l2_abc.B-l2_abc.A*l1_abc.B < 0) {
			// 小于零时到角计为 -90
			theta = -90;
		}
		if(!isZero(d)) {
			// 两直线不互相垂直，计算到角
			theta = Math.atan((l1_abc.A*l2_abc.B-l2_abc.A*l1_abc.B)/d)*180/Math.PI;
		}
		return theta;
	}
	/**
	 * 定义角平分线上一点。使用方法：tkzDefBisectorLine('A','B','C')('P')，在角 ABC 的平分线上取一点 P
	 * @param A: 三角形的顶点 A
	 * @param B: 三角形的顶点 B
	 * @param C: 三角形的顶点 C
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzDefBisectorLine = function(A, B, C) {
		// 取得到角
		let theta = this.tkzGetAngle(B,A,B,C);
		// 将 A 绕 B 旋转 theta/2
		return this.tkzDefRotaPoint(B, A, theta/2);
	}

	/**
	 * 取得两条直线的交点。使用方法：tkzInterLL('A,B,C,D')('X')，创建直线 AB,CD 的交点 X
	 * @param points_name: 分别为两条直线上的两点名称，例如 'A,B,C,D' 表示直线 AB, CD 的交点。
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzInterLL = function(points_name) {
		let point_names = points_name.split(',');
		let A1 = this.points[point_names[0]];
		let A2 = this.points[point_names[1]];
		let B1 = this.points[point_names[2]];
		let B2 = this.points[point_names[3]];
		let d = (A2.y-A1.y)*(B2.x-B1.x)-(A2.x-A1.x)*(B2.y-B1.y);
		let x = Infinity;
		let y = Infinity;
		if(!isZero(d)) {
			// 两直线不平行，有交点
			let m1 = B1.y*B2.x-B1.x*B2.y;
			let m2 = A1.y*A2.x-A1.x*A2.y;

			x = ((B1.x-B2.x)*m2-(A1.x-A2.x)*m1)/d;
			y = ((B1.y-B2.y)*m2-(A1.y-A2.y)*m1)/d;
		}
		this.tkz_tmp_points = [{'x': x, 'y': y}];
		return create_point;
	}
	/**
	 * 定义三角形的垂心。使用方法：tkzOrthoCenter('A','B','C')('H')，创建三角形 ABC 的垂心 H
	 * @param A: 三角形的顶点 A
	 * @param B: 三角形的顶点 B
	 * @param C: 三角形的顶点 C
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzOrthoCenter = function(A,B,C) {
		this.tkzDefProjPoint(B,C,A)('tkz_tmp_Xa');
		this.tkzDefProjPoint(A,B,C)('tkz_tmp_Xc');
		return this.tkzInterLL(`${A},tkz_tmp_Xa,${C},tkz_tmp_Xc`);
	}
	/**
	 * 定义三角形的外心。使用方法：tkzCircumCenter('A','B','C')('O')，创建三角形 ABC 的外心 O
	 * @param A: 三角形的顶点 A
	 * @param B: 三角形的顶点 B
	 * @param C: 三角形的顶点 C
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzCircumCenter = function(A,B,C) {
		this.tkzDefMidPoint(A,B)('tkz_tmp_Mc');
		this.tkzDefMidPoint(B,C)('tkz_tmp_Ma');
		this.tkzDefOrthLine(A,B,'tkz_tmp_Mc')('tkz_tmp_Xc');
		this.tkzDefOrthLine(B,C,'tkz_tmp_Ma')('tkz_tmp_Xa');
		return this.tkzInterLL('tkz_tmp_Ma,tkz_tmp_Xa,tkz_tmp_Mc,tkz_tmp_Xc');
	}
	/**
	 * 定义三角形的内心。使用方法：tkzInCenter('A','B','C')('I')，创建三角形 ABC 的内心 I
	 * @param A: 三角形的顶点 A
	 * @param B: 三角形的顶点 B
	 * @param C: 三角形的顶点 C
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzInCenter = function(A,B,C) {
		this.tkzDefBisectorLine(A,B,C)('tkz_tmp_Xb');
		this.tkzDefBisectorLine(B,C,A)('tkz_tmp_Xc');
		return this.tkzInterLL(`${B},tkz_tmp_Xb,${C},tkz_tmp_Xc`);
	}
	/**
	 * 定义三角形的重心。使用方法：tkzBaryCenter('A','B','C')('G')，创建三角形 ABC 的重心 G
	 * @param A: 三角形的顶点 A
	 * @param B: 三角形的顶点 B
	 * @param C: 三角形的顶点 C
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzBaryCenter = function(A,B,C) {
		this.tkzDefMidPoint(B,C)('tkz_tmp_Xa');
		this.tkzDefMidPoint(C,A)('tkz_tmp_Xb');
		return this.tkzInterLL(`${A},tkz_tmp_Xa,${B},tkz_tmp_Xb`);
	}
	/**
	 * 定义三角形的旁心。使用方法：tkzExCenter('A','B','C')('I_b')，创建三角形 ABC 关于角 B 的旁心 I_b
	 * @param A: 三角形的顶点 A
	 * @param B: 三角形的顶点 B
	 * @param C: 三角形的顶点 C
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzExCenter = function(A,B,C) {
		this.tkzDefBisectorLine(A,B,C)('tkz_tmp_Xb');
		this.tkzDefBisectorLine(B,C,A)('tkz_tmp_Xc');
		this.tkzDefRotaPoint(C,'tkz_tmp_Xc',90)('tkz_tmp_X');
		return this.tkzInterLL(`${C},tkz_tmp_X,${B},tkz_tmp_Xb`);
	}
	/**
	 * 取得直线与圆的交点。使用方法：tkzInterLC('A,B,O,P')('X,Y')，创建直线 AB 与以 O 为圆心 OP 为半径的圆的交点 X,Y。
	 * @param points_name: 直线上两点的名称和圆心名称和圆上一点的名称，例如 'A,B,O,P' 表示直线 AB, 圆 OP。
	 * @return: 返回创建点函数，创建两个点。
	 */
	this.tkzInterLC = function(points_name) {
		let point_names = points_name.split(',');
		let pA = this.points[point_names[0]];
		let pB = this.points[point_names[1]];
		let O = this.points[point_names[2]];
		let P = this.points[point_names[3]];
		// 圆半径
		let r = getR(O,P);
		// 直线一般方程的 A,B,C
		let abc = getABC(pA,pB);
		let m = abc.A*abc.A+abc.B*abc.B;

		let a2 = abc.A*abc.A;
		let b2 = abc.B*abc.B;
		let s1 = abc.C+abc.A*O.x;
		let s2 = abc.C+abc.B*O.y;

		let t = Math.sqrt(b2*(m*r*r-(s1+abc.B*O.y)*(s1+abc.B*O.y)));
		if(!isZero(abc.B)) {
			let x1 = (b2*O.x-abc.A*s2-t)/(a2+b2);
			let y1 = (-b2*s1+a2*abc.B*O.y+abc.A*t)/(abc.B*(a2+b2));
			let x2 = (b2*O.x-abc.A*s2+t)/(a2+b2);
			let y2 = (-b2*s1+a2*abc.B*O.y-abc.A*t)/(abc.B*(a2+b2));
			this.tkz_tmp_points = [{'x': x1, 'y': y1}, {'x': x2, 'y': y2}];
		} else {
			let x = -abc.C/abc.A;
			let p = Math.sqrt(-a2*s1*s1+a2*a2*r*r)/a2;
			let y1 = O.y-p;
			let y2 = O.y+p;
			this.tkz_tmp_points = [{'x': x, 'y': y1}, {'x': x, 'y': y2}];
		}
		return create_point;
	}
	/**
	 * 取得两圆的交点。使用方法：tkzInterCC('O1,A,O2,B')('X,Y')，创建圆 O1A 与圆 O2B 的交点 X,Y。
	 * @param points_name: 圆心名称和圆上一点的名称对，例如 'O1,A,O2,B' 表示圆 O1A 与圆 O2B。
	 * @return: 返回创建点函数，创建两个点。
	 */
	this.tkzInterCC = function(points_name) {
		let point_names = points_name.split(',');
		let O1 = this.points[point_names[0]];
		let A = this.points[point_names[1]];
		let O2 = this.points[point_names[2]];
		let B = this.points[point_names[3]];
		// 连心线长度
		let l = this.tkzCalcLength(O1.name,O2.name);
		// 圆 O1A 的半径
		let r1 = getR(O1,A);
		// 圆 O2B 的半径
		let r2 = getR(O2,B);
		// 设相交弦 P1P2 与连心线交于 M
		// 有 MP1^2 = r1^2-O1M^2 = r2^2-O2M^2，所以 r1^2-r2^2 = (O1M+O2M)(O1M-O2M) = l(O1M-O2M)
		// 计算 O1M 的长度
		let o1m = ((r1*r1-r2*r2)/l+l)/2;
		// 角 P1O1M
		let theta = Math.acos(o1m/r1)*180/Math.PI;

		let x1 = 0; let x2 = 0; let y1 = 0; let y2 = 0;

		// 将 O2 绕 O1 旋转 theta，得到点 X
		this.tkzDefRotaPoint(O1.name,O2.name,theta)('tkz_tmp_X');
		//直线 O1X 与圆 O2B 的交点 Y1, Y2
		this.tkzInterLC(`${O1.name},tkz_tmp_X,${O2.name},${B.name}`)('tkz_tmp_Y1,tkz_tmp_Y2');
		let Y1 = this.points['tkz_tmp_Y1'];
		let Y2 = this.points['tkz_tmp_Y2'];
		if(atCircle(O1,A,Y1) && atCircle(O2,B,Y1)) {
			// Y1 在两个圆上，是交点
			x1 = Y1.x;
			y1 = Y1.y;
		} else {
			// Y2 在两个圆上，是交点
			x1 = Y2.x;
			y1 = Y2.y;
		}
		// 将 O2 绕 O1 旋转 -theta，得到点 X
		this.tkzDefRotaPoint(O1.name,O2.name,-theta)('tkz_tmp_X');
		//直线 O1X 与圆 O2B 的交点
		this.tkzInterLC(`${O1.name},tkz_tmp_X,${O2.name},${B.name}`)('tkz_tmp_Z1,tkz_tmp_Z2');
		let Z1 = this.points['tkz_tmp_Z1'];
		let Z2 = this.points['tkz_tmp_Z2'];
		if(atCircle(O1,A,Z1) && atCircle(O2,B,Z1)) {
			// Z1 在两个圆上，是交点
			x2 = Z1.x;
			y2 = Z1.y;
		} else {
			// Z2 在两个圆上，是交点
			x2 = Z2.x;
			y2 = Z2.y;
		}
		this.tkz_tmp_points = [{'x': x1, 'y': y1}, {'x': x2, 'y': y2}];
		return create_point;
	}
	/**
	 * 过圆外一点作圆的切线，取得两个切点。使用方法：tkzDefTangentFrom('O', 'A', 'P')('X,Y')，过 P 作圆 OA 的切线，取得切点 X,Y。
	 * @param center_name: 圆心点名称
	 * @param point_name: 圆上一点名称
	 * @param p_name: 圆外一点名称
	 * @return: 返回创建点函数，创建两个点。
	 */
	this.tkzDefTangentFrom = function(center_name, point_name, p_name) {
		return this.tkzInterCC(`${center_name},${point_name},${p_name},${center_name}`);
	}
	/**
	 * 过圆上一点作圆的切线，并取得切线上一点。使用方法：tkzDefTangentAt('O', 'A')('X')，过 A 作圆 OA 的切线，取得切线上一点 X。
	 * @param center_name: 圆心点名称
	 * @param point_name: 圆上一点名称
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzDefTangentAt = function(center_name, point_name) {
		return this.tkzDefOrthLine(center_name, point_name, point_name);
	}
	/**
	 * 作两圆的内公切线，并取得切点。使用方法：tkzDefTangentInCC('O1', 'P1', 'O2', 'P2')('X1,X2,Y1,Y2')，作圆 O1P1 和圆 O2P2 的内公切线 X1X2, Y1Y2，切点为 X1, X2, Y1, Y2
	 * @param O1_name: 圆心 1 的名称
	 * @param P1_name: 圆 1 上一点名称
	 * @param O2_name: 圆心 2 的名称
	 * @param P2_name: 圆 2 上一点名称
	 * @return: 返回创建点函数，创建四个点。
	 */
	this.tkzDefTangentInCC = function(O1_name, P1_name, O2_name, P2_name) {
		//TODO 
	}
	/**
	 * 作点关于圆的反演点。使用方法：tkzDefInvPoint('O', 'A', 'P')('P'')，作点 P 关于圆 OA 的反演点 P'。
	 * @param center_name: 圆心点名称
	 * @param point_name: 圆上一点名称
	 * @param p_name: 被反演的点的名称
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzDefInvPoint = function(center_name, point_name, p_name) {
		let p = this.points[p_name];
		let op = this.tkzCalcLength(center_name, p_name);
		// 反演圆半径
		let r = this.tkzCalcLength(center_name, point_name);
		// 定比分点，OQ:QP = K;
		let m = op*op/r/r-1;
		if(!isZero(m)) {
			let K = 1/m;
			return this.tkzDefLinearPoint(center_name, p_name, K);
		} else {
			// 点在圆上，反演点即为自身。
			this.tkz_tmp_points = [{'x': p.x, 'y': p.y}];
			return create_point;
		}
	}

	/**
	 * 作调和第四点。使用方法：tkzDefHarmonicPoint('A', 'B', 'C')('D')，作点 D 使得 (AB,CD) = -1。
	 * @param A: 调和第一点的名称
	 * @param B: 调和第二点的名称
	 * @param C: 调和第三点的名称
	 * @return: 返回创建点函数，创建一个点。
	 */
	this.tkzDefHarmonicPoint = function(A_name, B_name, C_name) {
		let A = this.points[A_name];
		let B = this.points[B_name];
		let C = this.points[C_name];

		let abc = getABC(A,B);
		if(!isZero(abc.A*C.x+abc.B*C.y+abc.C)) {
			// 三点不共线，不能创建调和第四点
			this.tkz_msg = '三点不共线，不能创建调和第四点';
			return create_point_err;
		}
		let ac = this.tkzCalcLength(A_name, C_name, true);
		let bc = this.tkzCalcLength(B_name, C_name, true);
		// 调和点列：(AC:BC) = -(AD:BD)
		// 定比分点：K = AD:DB = -(DA:DB) = (AC:BC)
		let K = ac/bc;
		return this.tkzDefLinearPoint(A_name, B_name, K);
	}


	/**
	 * 绘制点
	 * @param params: 绘制参数，JSON 格式，例如 {color: 'black', fill: 'black', width: 2} 表示路径为黑色，填充为黑色，点宽为 2 像素
	 * @param points_name: 要绘制的点的字符串，以逗号分隔，例如 'A,B'
	 */
	this.tkzDrawPoints = function(params, points_name) {
		params = setParams(params);
		for(let point_name of points_name.split(',')) {
			let point = this.points[point_name];
			ctx.beginPath();
			ctx.strokeStyle = params.color;
			ctx.fillStyle = params.fill;
			ctx.arc(point.x * this.cm_pt, point.y * this.cm_pt, params.width, 0, 2 * Math.PI);
			ctx.fill();
			ctx.stroke();
		}
	}
	/**
	 * 绘制线段
	 * @param params: 绘制参数，JSON 格式，例如 {color: 'black', fill: 'black', width: 2, dash: true} 表示路径为黑色，填充为黑色，线宽为 2 像素，虚线
	 * @param start: 线段的起点
	 * @param to: 线段的终点
	 */
	this.tkzDrawSegment = function(params, start, to) {
		params = setParams(params);
		ctx.beginPath();
		ctx.lineCap = 'butt';
		ctx.lineWidth = params.width;
		ctx.strokeStyle = params.color;
		ctx.moveTo(start.x * this.cm_pt, start.y * this.cm_pt);
		ctx.lineTo(to.x * this.cm_pt, to.y * this.cm_pt);
		ctx.stroke();
	}
	/**
	 * 绘制线段
	 * @param params: 绘制参数，JSON 格式，例如 {color: 'black', fill: 'black', width: 2, dash: true} 表示路径为黑色，填充为黑色，线宽为 2 像素，虚线
	 * @param point_pairs_name: 线段的起点终点对字符串，起点与终点用逗号分隔，两个线段间用空格分隔，例如 'A,B C,D'
	 */
	this.tkzDrawSegments = function(params, point_pairs_name) {
		for(let point_pairs of point_pairs_name.split(' ')) {
			let start_to = point_pairs.split(',');
			let start = this.points[start_to[0]];
			let to = this.points[start_to[1]];
			this.tkzDrawSegment(params, start, to);
		}
	}
	/**
	 * 绘制圆
	 * @param params: 绘制参数，JSON 格式，例如 {color: 'black', width: 2, dash: true} 表示路径为黑色，线宽为 2 像素，虚线
	 * @param center_name: 圆心点字符串
	 * @param point_name: 圆上点字符串
	 */
	this.tkzDrawCircle = function(params, center_name, point_name) {
		params = setParams(params);
		let center = this.points[center_name];
		let point = this.points[point_name];
		ctx.beginPath();
		ctx.lineWidth = params.width;
		ctx.strokeStyle = params.color;
		ctx.arc(center.x * this.cm_pt, center.y * this.cm_pt, getR(center, point) * this.cm_pt, 0, 2 * Math.PI);
		ctx.stroke();
	}

	function getSubNumber(str) {
		if(str.indexOf('_') == -1) {
			return str;
		}
		s = str.match(/(.*)?_{(.*)}/);
		if(s && s.length != 3) {
			return str;
		}
		let new_str = s[1];
		for(let i = 0; i < s[2].length; i++) {
			var code = s[2].charCodeAt(i);
			if (code >= 48 && code <= 57) {
				new_str += String.fromCharCode(code + 8272);
			} else {
				new_str += s[2][i];
			}
		}
		return new_str;
	}
	/**
	 * 画出点名称
	 * @param params: 绘画参数
	 * @point_name: 点名称
	 */
	this.tkzLabelPoint = function(params, point_name) {
		params = setParams(params);
		let point = this.points[point_name];
		let p_name = getSubNumber(point.name);
		let x = point.x;
		let y = point.y;
		if(params.r && params.theta) {
			x += params.r * Math.cos(params.theta/180*Math.PI);
			y += -params.r * Math.sin(params.theta/180*Math.PI);
		}
		ctx.scale(1,-1);
		ctx.beginPath();
		ctx.fillStyle = params.color;
		ctx.font = params.font;
		ctx.fillText(p_name, x*this.cm_pt, -y*this.cm_pt);
		ctx.scale(1,-1);
	}
	/**
	 * 画出点序列名称。使用方法：tkzLabelPoints({color: 'blue'}, 'A,B') 画出 A,B 两点的标记，颜色为 blue
	 * @param params: 绘画参数
	 * @points_name: 点序列名称
	 */
	this.tkzLabelPoints = function(params, points_name) {
		points_name = points_name.split(',');
		for(let i = 0; i < points_name.length; i++) {
			this.tkzLabelPoint(params, points_name[i]);
		}
	}
	/**
	 * 画出角名称。使用方法：tkzLabelAngle({color: 'blue', label: 'θ', num: '20'}, 'A,B,C') 画出角 ABC 的标记，颜色为 blue，角名称为 θ，角度数为 20 度
	 * @param params: 绘画参数
	 * @points_name: 点序列名称
	 */
	this.tkzLabelAngle = function(params, points_name) {
		points_name = points_name.split(',');
		let A = this.points[points_name[0]];
		let B = this.points[points_name[1]];
		let C = this.points[points_name[2]];
		// TODO
	}
}
